@@ -3,7 +3,7 @@ import atexit
 import asyncio
 from dataclasses import dataclass
 from pathlib import Path
-from typing import Callable, Union
+from typing import Callable, Union, List
 from uuid import UUID
 
 from requests import exceptions
@@ -22,6 +22,15 @@ from progo.ble.constants import (
     UTF_8,
 )
 from progo.ble.util import find_gopros
+from progo.constants import (
+    OPEN_GOPRO_BASE_URL,
+    THUMBNAIL_DIR,
+    SCREENNAIL_DIR,
+    GPMF_DIR,
+    MEDIA_INFO_DIR,
+    TELEMETRY_DIR,
+    DCIM_DIR
+)
 from progo.exceptions import NoGoprosFoundException
 from progo.params import (
     FPS,
@@ -39,19 +48,54 @@ from progo.settings import SETTINGS_ATTR_INT_MAP, CameraSettings
 from progo.status import CameraStatus
 
 
-OPEN_GOPRO_BASE_URL = "http://10.5.5.9:8080/gopro"
-BASE_MEDIA_PATH = Path("100GOPRO/")
-
-
 def update_state(func):
     def wrap(self, *args, **kwargs):
         result = func(self, *args, **kwargs)
-        self._update_progo_state()
+        self._set_progo_state()
         return result
 
     return wrap
 
 
+def turbo_enabled(func):
+    def wrap(self, *args, **kwargs):
+        self.enable_turbo_mode()
+        result = func(self, *args, **kwargs)
+        self.disable_turbo_mode()
+        return result
+
+    return wrap
+
+
+@attr.s
+class MediaFile:
+    directory: str = attr.ib()
+    name: str = attr.ib()
+    created: int = attr.ib(converter=int)
+    modified: int = attr.ib(converter=int)
+    glrv: int = attr.ib()
+    ls: int = attr.ib()
+    size: int = attr.ib(converter=int)
+
+    @property
+    def camera_path(self):
+        return Path(self.directory, self.name)
+
+    @property
+    def gpmf_name(self):
+        """Add gpmf extension to the filename"""
+        return f"{self.name}.gpmf"
+
+    @property
+    def thumbnail_name(self):
+        """Add a jpg extension to the filename"""
+        return f"{self.name}.jpg"
+
+    @property
+    def screennail_name(self):
+        return self.thumbnail_name
+
+
 def run_async_function_synchronously(async_func: Callable, *args, **kwargs):
     """Because I want to use `bleak` but actively don't care about async."""
     coroutine = async_func(*args, **kwargs)
@@ -81,7 +125,6 @@ class Progo:
         self._ble_client: BleakClient = None
         self._wireless = Wireless()
         self._open_gopro_base_url = OPEN_GOPRO_BASE_URL
-        self._media_path = BASE_MEDIA_PATH
         self.status = None
         self.settings = None
         atexit.register(self._adios)
@@ -92,7 +135,7 @@ class Progo:
             run_async_function_synchronously(self._ble_client.disconnect)
 
     def _ensure_ble_connection(self):
-        logger.debug("Ensuring BLE connection")
+        logger.debug(f"Ensuring BLE connection to {self._gopro}")
         connected = self._ble_client.is_connected
         while not connected:
             logger.debug("BLE client not connected - connecting")
@@ -126,6 +169,16 @@ class Progo:
         response = requests.get(url, stream=True)
         return response
 
+    @turbo_enabled
+    def _http_get_contents_to_file(self, url: str, output_file: Path):
+        if output_file.exists():
+            logger.debug(f"{output_file} already exists - not downloading.")
+        else:
+            logger.debug(f"Writing {url} response contents to file {output_file}")
+            response = self._http_get(url)
+            output_file.parents[0].mkdir(parents=True, exist_ok=True)
+            output_file.write_bytes(response.content)
+
     def _get_json(self, url: str):
         response = self._http_get(url)
         return response.json()
@@ -171,12 +224,12 @@ class Progo:
     @update_state
     def set_third_party_client_info(self):
         # 2.0 only
-        return self._get_json("camera/analytics/set_client_info")
+        return self._get_json("gopro/camera/analytics/set_client_info")
 
     @update_state
     def start_encoding(self):
         # 2.0 only
-        return self._http_get("camera/shutter/start")
+        return self._http_get("gopro/camera/shutter/start")
 
     @update_state
     def set_shutter_on(self):
@@ -186,7 +239,7 @@ class Progo:
     @update_state
     def stop_encoding(self):
         # 2.0 only
-        return self._http_get("camera/shutter/stop")
+        return self._http_get("gopro/camera/shutter/stop")
 
     def set_shutter_off(self):
         # 2.0 only
@@ -194,79 +247,109 @@ class Progo:
 
     @update_state
     def set_digital_zoom(self, zoom_percent: int):
-        return self._get_json(f"camera/digital_zoom?percent={zoom_percent}")
+        return self._get_json(f"gopro/camera/digital_zoom?percent={zoom_percent}")
 
     def get_camera_state(self):
-        return self._get_json("camera/state")
+        return self._get_json("gopro/camera/state")
 
     def set_keep_alive(self):
-        return self._get_json("camera/keep_alive")
+        return self._get_json("gopro/camera/keep_alive")
 
-    def get_media_list(self):
-        return self._get_json("media/list")
+    def get_media_list(self) -> List[MediaFile]:
+        response = self._get_json("gopro/media/list")
+        media_list = []
+        for directory in response["media"]:
+            directory_name = directory["d"]
+            for file in directory["fs"]:
+                media_list.append(
+                    MediaFile(
+                        directory=directory_name,
+                        name=file["n"],
+                        created=file["cre"],
+                        modified=file["mod"],
+                        glrv=file.get("glrv"),
+                        ls=file.get("ls"),
+                        size=file["s"],
+                    )
+                )
+        return media_list
 
-    def download_file(self):
-        pass
-
-    def get_gpmf_data(self, filename: Path):
-        response = self._http_get(f"media/gpmf?path={self._media_path.join(filename)}")
-        # FIXME! Write response.data to a file locally
-
-    def get_media_info(self, filename: Path):
-        return self._http_get(f"media/info?path={self._media_path.join(filename)}")
-
-    def get_screennail(self, filename: str):
-        response = self._http_get(
-            f"media/screennail?path{self._media_path.join(filename)}"
+    def get_file(self, media_file: MediaFile, output_path: Path = None):
+        if not output_path:
+            output_path = DCIM_DIR.joinpath(media_file.name)
+        self._http_get_contents_to_file(
+            f"videos/DCIM/{media_file.camera_path}", output_path
         )
-        # FIXME! Write response.data to a file locally
 
-    def get_thumbnail(self, filename: str):
-        response = self._http_get(
-            f"media/thumbnail?path={self._media_path.join(filename)}"
+    def get_gpmf_data(self, media_file: MediaFile, output_path: Path = None):
+        if not output_path:
+            output_path = GPMF_DIR.joinpath(media_file.gpmf_name)
+        self._http_get_contents_to_file(
+            f"gopro/media/gpmf?path={media_file.camera_path}", output_path
         )
-        # FIXME! Write response.data to a file locally
 
-    def get_telemetry(self, filename: str):
-        response = self._http_get(
-            f"media/telemetry?path={self._media_path.join(filename)}"
+    def get_media_info(self, media_file: MediaFile, output_path: Path = None):
+        if not output_path:
+            output_path = MEDIA_INFO_DIR.joinpath(media_file.name)
+        self._http_get_contents_to_file(
+            f"gopro/media/info?path={media_file.camera_path}", output_path
+        )
+
+    def get_screennail(self, media_file: MediaFile, output_path: Path = None):
+        if not output_path:
+            output_path = SCREENNAIL_DIR.joinpath(media_file.screennail_name)
+        self._http_get_contents_to_file(
+            f"gopro/media/screennail?path={media_file.camera_path}", output_path
+        )
+
+    def get_thumbnail(self, media_file: MediaFile, output_path: Path = None):
+        if not output_path:
+            output_path = THUMBNAIL_DIR.joinpath(media_file.thumbnail_name)
+        self._http_get_contents_to_file(
+            f"gopro/media/thumbnail?path={media_file.camera_path}", output_path
+        )
+
+    def get_telemetry(self, media_file: MediaFile, output_path: Path = None):
+        if not output_path:
+            output_path = TELEMETRY_DIR.joinpath(media_file.telemetry_name)
+        self._http_get_contents_to_file(
+            f"gopro/media/telemetry?path={media_file.camera_path}", output_path
         )
-        # FIXME! Write response.data to a file locally
 
     def get_open_gopro_api_version(self):
-        return self._get_json("version")
+        return self._get_json("gopro/version")['version']
 
     def get_preset_status(self):
-        return self._get_json("camera/presets/get")
+        return self._get_json("gopro/camera/presets/get")
 
     @update_state
     def set_preset(self, preset: int) -> dict:
-        resp = self._get_json(f"camera/presets/load?id={preset}")
+        resp = self._get_json(f"gopro/camera/presets/load?id={preset}")
         return resp
 
     @update_state
     def set_preset_group(self, preset_group: PresetGroup) -> dict:
-        return self._get_json(f"camera/presets/set_group?id={preset_group.value}")
+        return self._get_json(f"gopro/camera/presets/set_group?id={preset_group.value}")
 
     @update_state
     def start_preview_stream(self):
-        return self._get_json("camera/stream/start")
+        return self._get_json("gopro/camera/stream/start")
 
     @update_state
     def stop_preview_stream(self):
-        return self._get_json("camera/stream/stop")
+        return self._get_json("gopro/camera/stream/stop")
 
     @update_state
     def enable_turbo_mode(self):
-        return self._get_json(f"media/turbo_transfer?p={Toggle.ON.value}")
+        return self._get_json(f"gopro/media/turbo_transfer?p={Toggle.ON.value}")
 
     @update_state
     def disable_turbo_mode(self):
-        return self._get_json(f"media/turbo_transfer?p={Toggle.OFF.value}")
+        return self._get_json(f"gopro/media/turbo_transfer?p={Toggle.OFF.value}")
 
     @update_state
     def _update_camera_setting(self, setting: int, option: int):
-        return self._get_json(f"camera/setting?setting={setting}&option={option}")
+        return self._get_json(f"gopro/camera/setting?setting={setting}&option={option}")
 
     def set_resolution(self, resolution: Resolution):
         resolution_id = SETTINGS_ATTR_INT_MAP["resolution"]
@@ -314,11 +397,10 @@ class Progo:
             connection_attempt += 1
         return connected
 
-    def _update_progo_state(self) -> bool:
-        max_attempts = 100
+    def _set_progo_state(self) -> bool:
         state_set = False
         attempt_num = 1
-        while not state_set and attempt_num <= max_attempts:
+        while not state_set:
             try:
                 state = self.get_camera_state()
                 self.settings = CameraSettings.from_dict(state["settings"])
@@ -336,5 +418,5 @@ class Progo:
         self._ensure_ble_connection()
         self.enable_wifi_ap()
         self.connect_to_wifi_ap()
-        self._update_progo_state()
+        self._set_progo_state()
         logger.info("Progo initialized")
